#!/usr/bin/env python
# encoding: utf-8

import ply.lex as lex
import ply.yacc as yacc
import re
import sys
import codecs

from wenlib import num_to_int

"""
This is the WEN (文) programming language.
Robert Östling, 2013-01-18

fundef = 為 expr 曰 sym
       | 以 var [var [...]] 為 expr 曰 sym
       | fundef 且 fundef
stmt = stmt 而 stmt
     | sym 者 expr 也
     | 答 expr
     | 印 expr
expr = var
     | num
     | 曰 string literal 已
     | 行 sym
     | 以 expr [及 expr [...]] 行 sym
     | 如 expr 則 expr 否 expr
     | expr 加/減/乘/分 expr
     | expr 齊 expr
     | 不 expr
"""

tokens = (
    'JI','WEI','ZHE','YE','YI','XING','ER','DA','WEI4',
    'YIN','QIUYAN','QIUSHU',
    'RU','ZE','FOU','QIE',
    'JIA','JIAN','CHENG','FEN','BU','QI',
    'SYM','NUM','STR')

def Lexer():
    t_JI        = ur'及'
    t_WEI       = ur'為'
    t_ZHE       = ur'者'
    t_YE        = ur'也'
    t_YI        = ur'以'
    t_XING      = ur'行'
    t_ER        = ur'而'
    t_QIE       = ur'且'
    t_WEI4      = ur'謂'
    t_DA        = ur'答'
    t_YIN       = ur'印'
    t_QIUYAN    = ur'求言'
    t_QIUSHU    = ur'求數'
    t_RU        = ur'如'
    t_ZE        = ur'則'
    t_FOU       = ur'否'
    t_JIA       = ur'加'
    t_JIAN      = ur'減'
    t_CHENG     = ur'乘'
    t_FEN       = ur'分'
    t_BU        = ur'不'
    t_QI        = ur'齊'

    def t_STR(t):
        ur'曰\w+?已'
        t.value = t.value[1:-1]
        return t

    def t_NUM(t):
        ur'[零一二三四五六七八九十百千萬億]+'
        t.value = num_to_int(t.value)
        return t

    t_SYM       = ur'[^曰已零一二三四五六七八九十百千萬億'+\
                  ur'及為者也以行而且謂答印求如則否加減乘分不齊]'

    t_ignore = u' \n\t'

    def t_error(t):
        print u"Illegal character '%s'" % t.value[0]
        t.lexer.skip()

    return lex.lex(reflags=re.UNICODE)

def Parser(lexer):
    def p_fundef_yiwei(p):
        'fundef : YI syms WEI stmt WEI4 SYM'
        p[0] = 'def f_%x(%s): %s' % (
            ord(p[6]),
            ','.join(['x_%x' % ord(arg) for arg in p[2]]),
            p[4])

    def p_fundef_wei(p):
        'fundef : WEI stmt WEI4 SYM'
        p[0] = 'def f_%x(): %s' % (ord(p[4]), p[2])

    def p_fundef_pair(p):
        'fundef : fundef QIE fundef'
        p[0] = p[1] + '\n\n' + p[3]

    def p_args_single(p):
        'args : expr'
        p[0] = [p[1]]

    def p_args_pair(p):
        'args : args JI expr'
        p[0] = p[1] + [p[3]]

    def p_expr_exec(p):
        'expr : YI args XING SYM'
        p[0] = 'f_%x(%s)' % (ord(p[4]), ','.join(p[2]))

    def p_expr_exec_noarg(p):
        'expr : XING SYM'
        p[0] = 'f_%x()' % ord(p[2])

    def p_expr_var(p):
        'expr : SYM'
        p[0] = 'x_%x' % ord(p[1])

    def p_expr_num(p):
        'expr : NUM'
        p[0] = str(p[1])

    def p_expr_str(p):
        'expr : STR'
        p[0] = 'u"' + p[1].encode('Raw-Unicode-Escape') + '"'

    def p_expr_if(p):
        'expr : RU expr ZE expr FOU expr'
        p[0] = '(%s) if (%s) else (%s)' % (p[4], p[2], p[6])

    def p_expr_bu(p):
        'expr : BU expr'
        p[0] = '(not (%s))' % p[2]

    def p_expr_qi(p):
        'expr : expr QI expr'
        p[0] = '((%s) == (%s))' % (p[1], p[3])

    def p_expr_jia(p):
        'expr : expr JIA expr'
        p[0] = '((%s) + (%s))' % (p[1], p[3])

    def p_expr_jian(p):
        'expr : expr JIAN expr'
        p[0] = '((%s) - (%s))' % (p[1], p[3])

    def p_expr_cheng(p):
        'expr : expr CHENG expr'
        p[0] = '((%s) * (%s))' % (p[1], p[3])

    def p_expr_fen(p):
        'expr : expr FEN expr'
        p[0] = '((%s) / (%s))' % (p[1], p[3])

    def p_stmt_qiuyan(p):
        'stmt : QIUYAN SYM'
        p[0] = 'x_%x = unicode(raw_input(),"utf-8")' % ord(p[2])

    def p_stmt_qiushu(p):
        'stmt : QIUSHU SYM'
        p[0] = 'x_%x = num_to_int(unicode(raw_input(),"utf-8"))' % ord(p[2])

    def p_stmt_print(p):
        'stmt : YIN expr'
        p[0] = 'print (%s)' % p[2]

    def p_stmt_seq(p):
        'stmt : stmt ER stmt'
        p[0] = p[1] + '; ' + p[3]

    def p_syms_single(p):
        'syms : SYM'
        p[0] = [p[1]]

    def p_syms_pair(p):
        'syms : syms SYM'
        p[0] = p[1] + [p[2]]

    def p_stmt_return(p):
        'stmt : DA expr'
        p[0] = 'return ' + p[2]

    def p_stmt_zheye(p):
        'stmt : SYM ZHE expr YE'
        p[0] = 'x_%x = %s' % (ord(p[1]), p[3])

    precedence = (
        ('left', 'QIE'),
        ('left', 'WEI', 'WEI4'),
        ('left', 'ER'),
        ('left', 'ZHE', 'YE'),
        ('left', 'YIN', 'QIUYAN', 'QIUSHU'),
        ('left', 'DA'),
        ('left', 'RU', 'ZE', 'FOU'),
        ('left', 'JIA', 'JIAN'),
        ('left', 'BU'),
        ('left', 'QI'),
        ('left', 'CHENG', 'FEN'),
        ('left', 'XING'),
    )

    def p_error(tok):
        raise SyntaxError, u"Parsing error: %s" % tok

    return yacc.yacc()

def main():
    if len(sys.argv) != 2:
        print >>sys.stderr, "wen.py file.wen >file.py && python file.py"
        exit(1)

    program = codecs.open(sys.argv[1], "rb", "utf-8").read()
    lexer = Lexer()
    parser = Parser(lexer)
    python = parser.parse(program)
    print 'from wenlib import *\n\n' + python + ('\n\nf_%x()' % ord(u'始'))

if __name__ == '__main__':
    main()

